local zapcolor = SKINCOLOR_CYAN

local sparktype = MT_SUPERSPARK
local zaptype = MT_WATERZAP

freeslot("S_PLAY_ROBEFALLTRANS")
states[S_PLAY_ROBEFALLTRANS] = {SPR_PLAY, SPR2_STND, 2, nil, 0, 0, S_PLAY_FALL}
freeslot("sfx_rbhovr", "sfx_rbstrt", "sfx_rbchrg", "sfx_rbdsh1", "sfx_rbdsh2", "sfx_rbdsh3", "sfx_rbregn", "sfx_rbbonk")
sfxinfo[sfx_rbhovr].caption = "Hovering"
sfxinfo[sfx_rbstrt].caption = "/"
sfxinfo[sfx_rbchrg].caption = "Charging"
sfxinfo[sfx_rbdsh1].caption = "/"
sfxinfo[sfx_rbdsh2].caption = "/"
sfxinfo[sfx_rbdsh3].caption = "Dashing"
sfxinfo[sfx_rbregn].caption = "Energy get"
sfxinfo[sfx_rbbonk].caption = "Clang"

// Robe's Energy Bar HUD
// Plus a very dangerous chasecam variable
local verydangerouschasecamvariable = true

hud.add(function(v, player, cam)
	verydangerouschasecamvariable = cam.chase

	if not player.mo or not player.mo.valid or player.mo.skin ~= "robe"
	or player.powers[pw_carry] & CR_NIGHTSMODE
		return
	end

	local mo = player.mo

	local i = 0

	while i <= 80 do
		local color = SKINCOLOR_JET
		local graphic = "ROBEBAR1"

		if not i
			graphic = "ROBEBAR2"
		elseif i == 80
			graphic = "ROBEBAR3"
		end

		if player.robeenergy and i <= player.robeenergy
			if player.powers[pw_super] > 1
				color = mo.color
			else
				color = zapcolor
			end
		end
		v.draw(i + 16, 166, v.cachePatch(graphic), V_HUDTRANS|V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT, v.getColormap(nil, color))
		i = $ + 1
	end
end)

// Spawns objects in a circle
local function robecircle(mobj, mobjtype, ns, zoffset, scale, keepmom, color, vertical, vertangle)
	local i = 0

	while i < 8 do
		local fa = i * (ANGLE_180/4)
		local mo = P_SpawnMobjFromMobj(mobj, 0, 0, zoffset, mobjtype)

		if not mo or not mo.valid
			continue
		end

		mo.scale = scale
		
		if color
			mo.color = color
			mo.colorized = true
		end

		if vertical
			mo.momz = FixedMul(sin(fa), ns) / 2
			P_InstaThrust(mo, vertangle + ANGLE_90, FixedMul(cos(fa), ns) / 2)
		else
			mo.momx = FixedMul(sin(fa), ns) / 2
			mo.momy = FixedMul(cos(fa), ns) / 2
		end

		if keepmom
			mo.momx = $ + mobj.momx
			mo.momy = $ + mobj.momy
			mo.momz = $ + mobj.momz
		end

		i = $ + 1
	end
end

// Followitem used for floating on the ground
freeslot("MT_ROBE_CLONE")
mobjinfo[MT_ROBE_CLONE] = {
	spawnstate = S_INVISIBLE, 
	spawnhealth = 1000, 
	radius = 16*FRACUNIT, 
	height = 48*FRACUNIT, 
	dispoffset = 0, 
	flags = MF_SCENERY|MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING
}

addHook("MobjSpawn", function(mobj)
	mobj.threshold = 0
	mobj.reactiontime = 0
	mobj.movecount = 1
end, MT_ROBE_CLONE)

addHook("FollowMobj", function(player, followmo)
	if not player.mo or not player.mo.valid
	or not followmo or not followmo.valid
		return true
	end

	local mo = player.mo
	local carry = player.powers[pw_carry]
	local state = mo.state

	if player.powers[pw_super] == 1 or player.guard
	or (carry == CR_NIGHTSMODE
	and mo.state >= S_PLAY_NIGHTS_TRANS1 and mo.state <= S_PLAY_NIGHTS_TRANS6)
		followmo.state = S_INVISIBLE
		return
	end

	local actLikeOnGround = P_IsObjectOnGround(mo)
		or mo.eflags & MFE_JUSTHITFLOOR
		or player.panim == PA_EDGE
		or player.panim == PA_RUN
		or player.robedash
		
	local spinning = player.pflags & PF_SPINNING
	local inStasis = player.pflags & PF_STASIS
	
	local speed = max(player.speed, 4*FRACUNIT)
	if carry
		speed = FRACUNIT/16
	end

	if carry == CR_MINECART or carry == CR_ROLLOUT
	or (actLikeOnGround and carry&~CR_NIGHTSMODE == CR_NONE
	and not spinning and not inStasis)
		followmo.reactiontime = $+speed
	elseif followmo.reactiontime > 90*FRACUNIT and followmo.reactiontime < 270*FRACUNIT
		followmo.reactiontime = (($-180*FRACUNIT)*3/4)+180*FRACUNIT
	elseif followmo.reactiontime == 180*FRACUNIT
		followmo.reactiontime = 0
	else
		followmo.reactiontime = $*3/4
	end
	
	if followmo.reactiontime >= 360*FRACUNIT
		followmo.reactiontime = $-360*FRACUNIT
	end

	local verticalflip = mo.eflags & MFE_VERTICALFLIP
	local floorz = mo.floorz
	local ceilingz = mo.ceilingz
	if floorz == ceilingz
		ceilingz = mo.floorz
	end

	if not verticalflip
		floorz = $ - mo.height
		ceilingz = $ - mo.height
	end
	
	local zoffset = FixedDiv(10 * sin(FixedAngle(followmo.reactiontime)), max(speed/10,FRACUNIT))
	local floorpos = max(mo.z + FixedMul(zoffset, followmo.scale), floorz)
	local zpos = min(floorpos, ceilingz)
	
	if SF_CANBUSTWALLS
		mo.spriteyoffset = zpos-mo.z
		followmo.state = S_INVISIBLE
		return
	end

	P_TeleportMove(followmo, mo.x, mo.y, zpos)
	
	followmo.skin = mo.skin
	followmo.state = mo.state
	followmo.frame = mo.frame
	followmo.angle = player.drawangle
	followmo.rollangle = mo.rollangle
	followmo.color = mo.color
	followmo.colorized = mo.colorized
	followmo.scale = mo.scale
	followmo.radius = mo.radius
	followmo.height = mo.height
	followmo.shadowscale = mo.shadowscale
	followmo.flags2 = mo.flags2 & (MF2_DONTDRAW|MF2_OBJECTFLIP)
	followmo.eflags = mo.eflags & MFE_VERTICALFLIP

	mo.flags2 = $ | MF2_DONTDRAW
	followmo.flags2 = $ & ~MF2_LINKDRAW

	return true
end, MT_ROBE_CLONE)

// Prevent turning in 2D and parse strafe controls to camera controls
addHook("PreThinkFrame", do
	for player in players.iterate do
		if player.mo and player.mo.valid
		and player.mo.skin == "robe" and player.robedash
			if (twodlevel or player.mo.flags2 & MF2_TWOD)
				player.cmd.forwardmove = 0
				player.cmd.sidemove = 0
			elseif not (player.pflags & PF_ANALOGMODE)
				local multiplier = 1

				if player != server
					multiplier = $*2
				end

				if player.cmd.sidemove != 0
					player.cmd.angleturn = $ + player.cmd.sidemove * (multiplier * -16)
					player.mo.angle = player.cmd.angleturn << 16
					player.drawangle = player.mo.angle
					player.cmd.sidemove = 0
				end
			end
		end
	end
end)

// Float ability
addHook("AbilitySpecial", function(player)
	if not player.mo or not player.mo.valid or player.mo.skin ~= "robe"
		return
	end

	if (player.pflags & (PF_THOKKED|PF_SHIELDABILITY)
	and not (player.charflags & SF_MULTIABILITY)) 
	or player.robefloat or not player.robeenergy
	or player.mo.frostburnt or player.dirk_frozen
	or player.iseggrobo or player.isjettysyn
	or player.gotflag or player.gotcrystal
	or player.guard or player.actionstate
		return true
	end
	
	if player.robedash
		if SF_CANBUSTWALLS
			player.charflags = $&~SF_CANBUSTWALLS
		else
			player.charability = CA_NONE
		end

		S_StopSoundByID(player.mo, sfx_rbdsh2)
		S_StopSoundByID(player.mo, sfx_rbdsh3)

		player.robedash = 0
	end

	player.mo.state = S_PLAY_FLY
	player.pflags = $ | PF_THOKKED & ~PF_SPINNING & ~PF_STARTJUMP
	player.robefloat = true
	
	//Slow down, but not while hyper
	if not (player.isxmomentum and player.powers[pw_super] and player.hyper
	and player.hyper.capable)
		player.normalspeed = skins[player.realmo.skin].normalspeed/2
		player.mo.momx = $ / 2
		player.mo.momy = $ / 2
		player.mo.momz = $ / 4
	end

	player.mo.momz = $ or P_GetMobjGravity(player.mo) * (P_IsObjectInGoop(player.mo) and -1 or 1)

	S_StartSound(player.mo, sfx_rbhovr)
	S_StartSound(player.mo, sfx_rbstrt)

	local thok
	if not SF_CANBUSTWALLS and player.followmobj and player.followmobj.valid
		player.followmobj.state = S_PLAY_FLY
		thok = P_SpawnGhostMobj(player.followmobj)
	else
		thok = P_SpawnGhostMobj(player.mo)
	end

	thok.destscale = player.mo.scale*8
	thok.scalespeed = $*3
	thok.color = SKINCOLOR_SUPERSKY1
	thok.colorized = true

	local angleturn = player.cmd.angleturn * FRACUNIT
	thok.momx = player.mo.momx - P_ReturnThrustX(nil, angleturn, player.mo.scale)
	thok.momy = player.mo.momy - P_ReturnThrustY(nil, angleturn, player.mo.scale)
	thok.momz = player.mo.momz

	return true
end)

// Charge ability
addHook("SpinSpecial", function(player)
	if not player.mo or not player.mo.valid or player.mo.skin ~= "robe"
		return
	end
	
	local mo = player.mo
	local cmd = player.cmd
	local shield = player.powers[pw_shield] & SH_NOSTACK

	if P_SuperReady(player)
	or player.pflags & PF_SPINDOWN
		return
	elseif player.powers[pw_super] == 1
	or not player.robeenergy
	or (P_PlayerInPain(player) or player.panim == PA_PAIN)
	or player.powers[pw_carry]
	or player.spectator or player.exiting
	or player.playerstate != PST_LIVE or not player.mo.health
	or player.pflags & (PF_SLIDING|PF_SHIELDABILITY)
	or player.powers[pw_nocontrol]
	or player.iseggrobo or player.isjettysyn
	or player.guard or player.actionstate
	or player.gotflag or player.gotcrystal
	or player.mo.frostburnt or player.dirk_frozen
	or (shield and shield != SH_PITY and shield != SH_PINK // Has shield and it's not pity or amy?
	and cmd.buttons & BT_JUMP and not (player.pflags & PF_SHIELDABILITY)) // Pressing jump and can use shield?
		return true
	end
	
	if player.robefloat
		player.robefloat = false
		player.normalspeed = skins[player.realmo.skin].normalspeed
		S_StopSoundByID(player.realmo, sfx_rbhovr)
	end

	if not P_IsObjectOnGround(mo)
		player.pflags = $ | PF_THOKKED
	end

	player.pflags = $ & ~PF_SPINNING
	player.robedash = 1
	
	if player.isxmomentum
		player.robedashspeed = max(player.speed*5/3, player.normalspeed*5/3)
		//Hyper version
		if player.powers[pw_super] and player.hyper and player.hyper.capable
			player.robedashspeed = $*3/2
		end
	elseif player.powers[pw_super] or player.powers[pw_sneakers]
		player.robedashspeed = player.normalspeed*5/2
	else
		player.robedashspeed = player.normalspeed*5/3
	end

	S_StartSound(player.mo, sfx_rbchrg)
	mo.state = S_PLAY_GLIDE

	local thok
	if not SF_CANBUSTWALLS and player.followmobj and player.followmobj.valid
		player.followmobj.state = S_PLAY_GLIDE
		thok = P_SpawnGhostMobj(player.followmobj)
	else
		thok = P_SpawnGhostMobj(player.mo)
	end

	thok.scale = mo.scale*3
	thok.z = $-(mo.height*3/2)
	thok.destscale = mo.scale
	thok.scalespeed = $*3
	thok.color = SKINCOLOR_SUPERSKY1
	thok.colorized = true

	local angleturn = cmd.angleturn * FRACUNIT
	thok.momx = mo.momx * 2/3 - P_ReturnThrustX(nil, angleturn, mo.scale)
	thok.momy = mo.momy * 2/3 - P_ReturnThrustY(nil, angleturn, mo.scale)
	thok.momz = 0

	return true
end)

// Pterabyte is a pun on the words terabyte and pterodactyl.
// A Terabyte is an extremely large unit of information often used with electronic devices.
// It is equal to one million million bytes, and is considered a level up from a gigabyte.
// It makes sense why this enemy's name would create a pun out of that word, 
// because the enemy itself is a robot, which is a electronic device, which as
// stated before are known to use these types of units of information.
// The fact that the name is a pun of the word pterodactyl as well makes sense
// due to the fact that the robot itself is modelled after the species of the
// genus of dinosaur of the same name. Its head shape, large wings, and
// bird-like feet with claws are characteristic of the animal, which this
// robotic replica captures quite wonderfully.
// Since the p in pterodactyl is silent, it can be assumed that the p in
// Pterabyte is silent as well, which makes its pronouncation the exact same
// as terabyte, which makes the pun even more pronounced and more clever.
// Anyone can say with complete confidence that this is objectively a good pun.
// Oh yeah, and this code murders them when they try to grab you.
addHook("TouchSpecial", function(special, toucher)
	if toucher.player and toucher.skin == "robe"
	and (toucher.player.robedash or toucher.player.robefloat)
		P_DamageMobj(special, toucher)
	end
end, MT_PTERABYTE)

// Bonking into a wall during the charge
addHook("MobjMoveBlocked", function(mo)
	if not mo or not mo.valid or mo.skin ~= "robe"
	or not mo.player or mo.player.robedash <= 15
		return
	end

	local player = mo.player

	P_BounceMove(mo)

	local momangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	local radius = 16*FRACUNIT
	local zoffset = mo.height/2

	robecircle(mo, sparktype, radius, zoffset, mo.scale, false, SKINCOLOR_NONE, true, InvAngle(momangle))

	if player.pflags & PF_ANALOGMODE and not (twodlevel or mo.flags2 & MF2_TWOD)
		player.drawangle = momangle
	else
		mo.angle = momangle
	end

	player.pflags = $ | PF_STASIS
	player.robedashspeed = $ / 4

	S_StartSound(mo, sfx_rbbonk)
end, MT_PLAYER)

// Destroy spikes by ramming into them
local function waterspike(spike, mo)
	// Return if neither exist.
	// Must be robe and robedash must be >15
	if not spike or not spike.valid
	or not mo or not mo.valid
	or not mo.player or not mo.skin == "robe"
	or not (mo.player.robedash > 15)
		return
	end
	
	// Within the correct height restraints? Kill the spike!
	if spike.z - spike.scale <= mo.z + mo.height
	and spike.z + spike.height + spike.scale >= mo.z
		P_KillMobj(spike, mo)
		return false
	end
end

addHook("MobjCollide", waterspike, MT_SPIKE)
addHook("MobjCollide", waterspike, MT_WALLSPIKE)

// No using shields if you're not holding jump at the same time
addHook("ShieldSpecial", function(player)
	if not player.mo or not player.mo.valid
	or player.mo.skin ~= "robe"
		return
	end

	local shield = player.powers[pw_shield] & SH_NOSTACK

	if (shield == SH_PITY or shield == SH_PINK
	or not (player.cmd.buttons & BT_JUMP)
	or player.pflags & PF_SHIELDABILITY
	or player.robedash or player.robefloat)
	and player.robeenergy
		return true
	end
end)

// Allow the abilities to damage
addHook("PlayerCanDamage", function(player, target)
	if not player.mo or not player.mo.valid
	or player.mo.skin ~= "robe"
		return
	end

	if player.robedash or player.robefloat
		return true
	end
end)

// Get energy back from enemies
addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
	if not inflictor or not inflictor.valid or inflictor.skin ~= "robe"
	or not inflictor.player
		return
	end

	local player = inflictor.player

	if player.robeenergy + (player.roberegen*3) >= 80
	or not target or not target.valid or not (target.flags & (MF_ENEMY|MF_BOSS))
		return
	end

	player.roberegen = $ + 8
	S_StartSound(nil, sfx_rbregn, player)

	if not SF_CANBUSTWALLS and player.followmobj and player.followmobj.valid
		P_SpawnGhostMobj(player.followmobj)
	else
		P_SpawnGhostMobj(player.mo)
	end
end)

// Electricity? Spikes? Damage floors? HA!
addHook("ShouldDamage", function(target, inflictor, source, damage, damagetype)
	if not target or not target.valid or target.skin ~= "robe"
	or not target.player or target.player.powers[pw_carry] & CR_NIGHTSMODE
		return
	end

	local inflictorvalid = inflictor and inflictor.valid
	local sourcevalid = source and source.valid
	
	//Your lightning only makes me more powerful!
	if damagetype == DMG_ELECTRIC
		//Hack to damage Brak even with his barrier up
		if inflictorvalid
			local brak = inflictor.target
			if inflictor.type == MT_CYBRAKDEMON_ELECTRIC_BARRIER
			and brak and brak.valid and brak.type == MT_CYBRAKDEMON
			and brak.health and not (brak.flags2 & MF2_FRET)
			and not (brak.flags & (MF_SPECIAL|MF_SHOOTABLE))
			and P_PlayerCanDamage(player, brak)
				brak.flags = $|MF_SPECIAL|MF_SHOOTABLE
				
				//Moving in place redoes object collision with the new flags we gave
				P_TryMove(target, target.x, target.y, true)
				
				brak.flags = $&~MF_SPECIAL&~MF_SHOOTABLE
			end
		end
		
		return false
	end

	// Spikes below us? What are those?
	if inflictorvalid
	and inflictor.type == MT_SPIKE
	and inflictor.eflags & MFE_VERTICALFLIP == target.eflags & MFE_VERTICALFLIP
		return false
	end

	// Sector fire or spike damage? Nope.
	if target.eflags & MFE_ONGROUND and P_IsObjectOnGround(target) // On ground?
	and not inflictorvalid // No inflictor?
	and not sourcevalid // No source?
	// Fire or spike damage?
	and ((damagetype == DMG_FIRE and not (target.eflags&MFE_UNDERWATER))
	or damagetype == DMG_SPIKE)
		return false
	end
end, MT_PLAYER)

// Reset Robe's variables
local function resetrobe(player)
	if player.robefloat
		player.normalspeed = skins[player.realmo.skin].normalspeed
		S_StopSoundByID(player.realmo, sfx_rbhovr)
	end

	player.robefloat = false

	if player.robedash
		if player.mo.state == S_PLAY_DASH
		or player.mo.state == S_PLAY_GLIDE
			if P_IsObjectOnGround(player.mo)
				player.mo.state = S_PLAY_RUN
			else
				player.mo.state = S_PLAY_ROBEFALLTRANS
			end
		end

		if SF_CANBUSTWALLS
			player.charflags = $&~SF_CANBUSTWALLS
		else
			player.charability = CA_NONE
		end

		S_StopSoundByID(player.realmo, sfx_rbdsh2)
		S_StopSoundByID(player.realmo, sfx_rbdsh3)
	end

	player.robedashspeed = player.normalspeed*5/3
	player.robedash = 0
end

addHook("PlayerSpawn", function(player)
	resetrobe(player)
	player.robeenergy = 80
	player.roberegen = 0
end)

// Main Thinker	
addHook("PlayerThink", function(player)
	if not player.mo or not player.mo.valid
	or player.mo.skin ~= "robe"
		return
	end
	
	local mo = player.mo
	local cmd = player.cmd
	local flip = P_MobjFlip(mo)
	local moGravity = P_GetMobjGravity(mo)
	local onGround = P_IsObjectOnGround(mo)
	local in2D = twodlevel or mo.flags2 & MF2_TWOD

	// You need to hold jump or fire to turn super
	if cmd.buttons & (BT_JUMP|BT_ATTACK)
		if not player.robedash and not player.robefloat
			player.charflags = $ | SF_SUPER
		end
	elseif not player.powers[pw_super] and player.charflags & SF_SUPER
		player.charflags = $ & ~SF_SUPER
	end
	
	// You need to hold jump to use shields
	if cmd.buttons & BT_JUMP and SF_NOSHIELDABILITY 
		player.charflags = $ &~ SF_NOSHIELDABILITY 
	elseif SF_NOSHIELDABILITY 
		player.charflags = $ | SF_NOSHIELDABILITY
	end
	
	// Energy regeneration
	if not (player.pflags & PF_THOKKED)
	and not player.robedash and not player.robefloat
	and mo.state != S_PLAY_GLIDE
		if player.robeenergy == nil
			player.robeenergy = 80
		//Energy regen is slower when not on the ground
		elseif player.robeenergy < 80
			if not(CBW_Battle)
				player.robeenergy = min($ + 4, 80)
			elseif (onGround or (onSurface and tracerOnGround) or not (leveltime%4))
				player.robeenergy = $ + 1
			end
		end
	elseif not mo.health
		player.robeenergy = max($-2, 0)
	end

	if player.powers[pw_super]
		if player.robeenergy == nil
			player.robeenergy = 80
		elseif player.robeenergy < 80
			player.robeenergy = $ + 1
		end
	end

	if player.robeprevrings != player.rings
		if player.robeprevrings != nil
		and player.robeprevrings < player.rings
			local deltarings = player.rings - player.robeprevrings

			player.roberegen = min($ + deltarings, 80 - (player.robeenergy/3))
		end

		player.robeprevrings = player.rings
	end

	if player.roberegen
		player.robeenergy = min($ + 2, 80)
		player.roberegen = $-1
	end

	// Disallows the rest of the code in certain circumstances
	if player.powers[pw_super] == 1
	or P_PlayerInPain(player) or player.panim == PA_PAIN
	or player.powers[pw_carry]
	or player.spectator or player.exiting
	or player.playerstate != PST_LIVE or not mo.health
	or player.pflags & PF_SLIDING or player.powers[pw_nocontrol]
	or player.iseggrobo or player.isjettysyn
	or player.guard or player.actionstate
		resetrobe(player)
		return
	end
	
	// Jump to fall transition state
	if (mo.state == S_PLAY_JUMP or mo.state == S_PLAY_SPRING
	or mo.state == S_PLAY_WALK) and not P_IsObjectOnGround(player.mo)
	and mo.momz * flip <= (-2 * moGravity) * flip
		mo.state = S_PLAY_ROBEFALLTRANS
	end
	
	// Always have the jump flag when in the air
	if not onGround and not (player.pflags & PF_JUMPED)
		player.pflags = $ | PF_JUMPED
	end
	
	// Dashing
	if player.robedash
		player.robedash = $ + 1

		if mo.momz * flip < 0
			mo.momz = 0
		end
		
		if not P_IsObjectOnGround(mo) and not (player.pflags & PF_THOKKED)
			player.pflags = $ | PF_THOKKED
		end

		if player.robedash < 15
			if mo.state != S_PLAY_GLIDE
				mo.state = S_PLAY_GLIDE
			end

			if player.robedash < 5 and player.robeenergy
				if not(CBW_Battle)
					player.robeenergy = $ - 1
				else
					player.robeenergy = max($-2, 0)
				end
			end

			player.pflags = $ | PF_FULLSTASIS

			mo.momx = FixedMul($, FRACUNIT * 94/100)
			mo.momy = FixedMul($, FRACUNIT * 94/100)
		elseif (player.gotflag or player.gotcrystal)
		or not (cmd.buttons & BT_SPIN) or not player.robeenergy
			if SF_CANBUSTWALLS
				player.charflags = $&~SF_CANBUSTWALLS
			else
				player.charability = CA_NONE
			end

			S_StopSoundByID(mo, sfx_rbdsh2)
			S_StopSoundByID(mo, sfx_rbdsh3)

			player.robedash = 0

			if onGround
				mo.state = S_PLAY_RUN
			else
				player.pflags = $ | PF_JUMPED
				mo.state = S_PLAY_ROBEFALLTRANS
			end

			mo.momx = $*2/3
			mo.momy = $*2/3
		else
			mo.state = S_PLAY_DASH

			if player.robedash == 15
				if SF_CANBUSTWALLS
					player.charflags = $|SF_CANBUSTWALLS
				else
					player.charability = CA_GLIDEANDCLIMB
				end

				S_StartSound(mo, sfx_rbdsh1)
				S_StartSound(mo, sfx_rbdsh2)

				robecircle(mo, sparktype, 64*FRACUNIT, mo.height/2, mo.scale, true, zapcolor)
				P_RadiusAttack(mo, mo, RING_DIST, DMG_ELECTRIC)
			end

			if not S_SoundPlaying(mo, sfx_rbdsh3)
				S_StartSound(mo, sfx_rbdsh3)
			end

			if player.robedashspeed != player.normalspeed*5/3
				if player.robedashspeed < 20*FRACUNIT
					player.pflags = $ | PF_STASIS
				end
				
				if player.robedashspeed > skins[player.realmo.skin].normalspeed*5/3
					player.robedashspeed = $ + FRACUNIT/4
				else
					player.robedashspeed = $ + 2*FRACUNIT
				end
				
				if player.isxmomentum
					player.normalspeed = max($, player.robedashspeed)
				elseif player.powers[pw_super] or player.powers[pw_sneakers]
					player.robedashspeed = min($, player.normalspeed*5/2)
				else
					player.robedashspeed = min($, player.normalspeed*5/3)
				end
			end

			if player.pflags & PF_ANALOGMODE and not in2D
				P_InstaThrust(mo, player.drawangle, FixedMul(player.robedashspeed, mo.scale))
			else
				P_InstaThrust(mo, mo.angle, FixedMul(player.robedashspeed, mo.scale))
			end

			// Make the dash usable for rising Spindashable platforms
			if mo.eflags & MFE_ONGROUND
				local sectorthing = P_PlayerTouchingSectorSpecial(player, 0, 0)

				local fofSpecials = {150, 151, 152, 153, 176, 177, 180, 190, 191, 192, 193, 194, 195}

				if sectorthing and sectorthing.valid
					for fof in sectorthing.ffloors()
						if not fof or not fof.valid
						or not fof.master or not fof.master.valid
							continue
						end

						if (abs(fof.topheight - mo.z) <= 40*mo.scale
						or abs(fof.bottomheight - mo.z + mo.height) <= 40*mo.scale)
						and fof.master.flags & ML_NOCLIMB
							local fofHasSpecial = false

							for i, v in ipairs(fofSpecials) do
								if fof.master.special == v
									fofHasSpecial = true
									break
								end
							end

							if not fofHasSpecial
								continue
							end

							player.pflags = $ | PF_STARTDASH

							P_InstaThrust(mo, 0, 0)
						end
					end
				end
			end

			player.robeenergy = $ - 1
		end
	end
	
	// Floating
	if player.robefloat
		// Cancel
		if not (cmd.buttons & BT_JUMP)
		or player.gotflag or player.gotcrystal
		or not player.robeenergy or onGround
			player.robefloat = false
			mo.momz = $*2/3
			player.normalspeed = skins[player.realmo.skin].normalspeed
			S_StopSoundByID(player.realmo, sfx_rbhovr)
			mo.state = S_PLAY_SPRING
		// Go up
		else
			if not S_SoundPlaying(mo, sfx_rbhovr)
				S_StartSound(mo, sfx_rbhovr)
			end
				
			//Hyper version
			if player.isxmomentum and player.powers[pw_super] and player.hyper
			and player.hyper.capable and mo.momz*flip < 32*mo.scale
				mo.momz = $ + abs(moGravity) * 8 * flip
				mo.momz = max(min(mo.momz, 32*mo.scale), -32*mo.scale)
			elseif mo.momz*flip < 12*mo.scale
				mo.momz = $ + abs(moGravity) * 4 * flip
				mo.momz = max(min(mo.momz, 12*mo.scale), -12*mo.scale)
			end

			mo.momz = $ or flip

			player.pflags = $ | PF_CANCARRY
			if not (CBW_Battle) or (player.robeenergy < 20)
				player.robeenergy = $ - 1
			else
				player.robeenergy = $ - 2
			end
			mo.state = S_PLAY_FLY
		end
	end
	
	// Prevent the electric effect and Ring/Enemy attraction
	if (not player.robefloat and not player.robedash)
	or player.powers[pw_flashing] >= TICRATE*2
		return
	end

	// Ring/Enemy attraction
	local maxdist = FixedMul(RING_DIST*2/3, mo.scale)
	local span = ANGLE_45
	local thokmobj = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
	thokmobj.radius = maxdist

	local rings = {
		MT_RING, MT_FLINGRING, MT_COIN, MT_FLINGCOIN,
		MT_BLUESPHERE, MT_FLINGBLUESPHERE
	}

	searchBlockmap("objects", function(refmobj, object)
		if object == mo
		or object.flags & (MF_NOCLIPTHING|MF_MISSILE)
		or object.flags2 & MF2_FRET
		or object.health <= 0
		or not P_CheckSight(mo, object)
		or (in2D	and abs(mo.y - object.y) > mo.radius)
			return
		end

		local isRing = false // Is a ring?

		// Iterate ring types.
		for i, v in ipairs(rings) do
			if object.type == v // We're a ring type?
				isRing = true // Set isRing and break
				break
			end
		end

		// Not a ring or our shield is an attraction shield.
		if (not isRing or (player.powers[pw_shield] & SH_NOSTACK) == SH_ATTRACT)
		and (not (object.flags & MF_ENEMY)
		or object.info.spawnhealth > 1
		or (object.type == MT_EGGGUARD and object.tracer)
		or (object.type == MT_BIGMINE and object.state >= S_BIGMINE_SET1)
		or (object.type == MT_BUMBLEBORE and object.state >= S_BUMBLEBORE_RAISE))
			return
		end

		local zdist = (mo.z + mo.height/2) - (object.z + mo.height/2)
		local dist = FixedHypot(mo.x - object.x, mo.y - object.y)

		if R_PointToAngle2(0, 0, dist, zdist) + span > span*2
		and R_PointToAngle2(0, 0, dist, -1*zdist) + span > span*2
			return // Don't home outside of desired angle!
		end
		
		dist = FixedHypot(dist, zdist*2)

		if (dist > maxdist)
			return // out of range
		end
		
		if isRing and object.info.reactiontime
		and object.type != object.info.reactiontime
			local newring = P_SpawnMobjFromMobj(object, 0, 0, 0, object.info.reactiontime)
			P_RemoveMobj(object)
			object = newring
		end
		
		local speed = object.info.speed
		object.info.speed = 80*FRACUNIT

		P_HomingAttack(object, mo)
		object.info.speed = speed
	end, thokmobj)

	P_RemoveMobj(thokmobj)
	
	// Electric effect
	local thok = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
	thok.target = mo
	thok.state = S_CYBRAKDEMONELECTRICBARRIER1 + P_RandomKey(24)
	thok.scale = mo.scale/2
	thok.fuse = 2

	// Only every four frames
	if leveltime % 4
		return
	end

	if player.robefloat or player.robedash < 15
		local thok = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
		thok.scale = mo.scale * 5/4
		thok.z = $ - (mo.height / 4)
		thok.destscale = mo.scale / 16
		thok.color = zapcolor
		thok.momx = mo.momx-P_ReturnThrustX(nil, cmd.angleturn * FRACUNIT, mo.scale)
		thok.momy = mo.momy-P_ReturnThrustY(nil, cmd.angleturn * FRACUNIT, mo.scale)
		thok.momz = mo.momz
	end

	if player.robefloat
		local speed = 2*FRACUNIT

		P_SetObjectMomZ(mo, speed, true)
		robecircle(mo, sparktype, 8 * FRACUNIT, 0, mo.scale/4, true, zapcolor)
		P_SetObjectMomZ(mo, -speed, true)
		robecircle(mo, zaptype, 20 * FRACUNIT, mo.height/2, mo.scale, true, zapcolor)
	end

	if player.robedash <= 15
		return
	end
	
	local thok
	if not SF_CANBUSTWALLS and player.followmobj and player.followmobj.valid
		thok = P_SpawnGhostMobj(player.followmobj)
	else
		thok = P_SpawnGhostMobj(player.mo)
	end
	
	thok.color = SKINCOLOR_SUPERSKY1
	thok.colorized = true
	thok.flags2 = $ | MF2_SHADOW

	robecircle(mo, zaptype, 18*FRACUNIT, mo.height/2, mo.scale, true, zapcolor, true, player.drawangle)

	for i = 0, 10 do
		local spark = P_SpawnMobjFromMobj(mo,
			50 * (P_RandomFixed() - FRACUNIT/2) + mo.momx, 
			50 * (P_RandomFixed() - FRACUNIT/2) + mo.momy, 
			50 * (P_RandomFixed() - FRACUNIT/2) + mo.height/2,
			sparktype
		)

		if not spark or not spark.valid
			continue
		end

		spark.momx = mo.momx/2
		spark.momy = mo.momy/2
		// spark.momz = mo.momz/2

		spark.scale = $/2
		spark.destscale = spark.scale/2
		spark.color = zapcolor
		spark.colorized = true

		if not verydangerouschasecamvariable
		and player == displayplayer
			spark.flags2 = $ | MF2_DONTDRAW
		end
	end
end)

//Battle Mod Support
local ScriptLoaded = false
addHook("ThinkFrame", function()
	if not(CBW_Battle) or ScriptLoaded return end
	ScriptLoaded = true
	
	local function GarbagePriority4(player)
		local checkshield = function(shield, p)
			return p.pflags&PF_SHIELDABILITY and (p.powers[pw_shield] & SH_NOSTACK) == shield
		end
		
		if player.robedash > 15
			CBW_Battle.SetPriority(player,1,0,"amy_melee",2,1,"electric vortex")
		elseif player.pflags&PF_JUMPED and not player.homing and not player.robefloat
		and not (player.pflags&(PF_NOJUMPDAMAGE|PF_SPINNING|PF_GLIDING))
		and not checkshield(SH_ELEMENTAL, player) and not checkshield(SH_BUBBLEWRAP, player)
		and not checkshield(SH_FLAMEAURA, player) and not checkshield(SH_THUNDERCOIN, player)
		and player.mo.state != S_PLAY_ROLL
			CBW_Battle.SetPriority(player,1,0,nil,1,0,"jumping spin attack")
		end
	end
	
	CBW_Battle.SkinVars["robe"] = {
		weight = 50,
		func_priority_ext = GarbagePriority4
	}
end)